import numpy as np
import multiprocessing as mp

from scipy.stats import entropy
from .utils import p_joint, mi_x1x2_c, compute_upper_bound

np.seterr(divide='ignore', invalid='ignore')


def empirical_bottleneck(x,y,return_entropies=False,numuniquex=0,numuniquey=0,**kwargs):
    """ Compute an IB curve for two empirical sequences x and y - main function
    
    Arguments:
    x -- first empirical sequence ("past")
    y -- second empirical sequence ("future")
    return_entropies (bool) -- whether to return the marginal entropies of x and y
    kwargs -- additional keyword arguments to be passed to IB().
    
    Returns:
    i_p -- values of ipast for each value of beta
    i_f -- values of ifuture for each value of beta
    beta -- values of beta considered
    mixy -- mutual information between x and y (curve saturation point) (only returned if return_entropies is True)
    hx -- entropy of x (only returned if return_entropies is True)
    hy -- entropy of y (only returned if return_entropies is True)
    """
    
    # Marginal, joint and conditional distributions required to calculate the IB
    pxy_j = p_joint(x,y)
    px = pxy_j.sum(axis=1)
    py = pxy_j.sum(axis=0)
    pyx_c = pxy_j.T / px
    
    #Calculate the information bottleneck for a range of values of beta
    i_x,i_y,beta,mixy,hx = IB(px,py,pyx_c,**kwargs)
    hy = entropy(py,base=2)
    
    # Return array of ipasts and ifutures for array of different values of beta - mixy should correspond to the saturation point
    if return_entropies:
        return i_x, i_y, beta, mixy, hx, hy
    else:
        return i_x, i_y, beta
    
def beta_iter(b,px,py,pyx_c,pm_size,restarts,iterations):
    """Function to run BA algorithm for individual values of beta
    
    Arguments:
    b -- value of beta on which to run algorithm
    px -- marginal probability distribution for the past (x)
    py -- marginal probability distribution for the future (y)
    pyx_c -- conditional distribution p(y|x)
    pm_size -- discrete size of the compression distribution
    restarts -- number of times the optimization procedure should be restarted (for each value of beta) from different random initial conditions
    iterations -- maximum number of iterations to use until convergence
    
    Returns:
    list with i_p and i_f, which correspond to ipast and ifuture values for each value of beta
    """
    candidates = []
    for r in range(restarts):
    
	    # Initialize distribution for bottleneck variable
	    pm = np.random.rand(pm_size)+1
	    pm /= pm.sum()                  
	    pym_c = np.random.rand(py.size,pm.size)+1 # Starting point for the algorithm
	    pym_c /= pym_c.sum(axis=0)
	    
	    # Iterate the BA algorithm
	    for i in range(iterations):
		    pmx_c, z = p_mx_c(pm,px,py,pyx_c,pym_c,b)
		    pm = p_m(pmx_c,px)
		    pym_c = p_ym_c(pm,px,py,pyx_c,pmx_c)
		    if i>0 and np.allclose(pmx_c,pmx_c_old,rtol=1e-3,atol=1e-3):
				# if the x->m mapping is not updating any more, we're at convergence and we can stop
			    break
		    pmx_c_old = pmx_c
	    candidates.append({'past_info' : mi_x1x2_c(pm, px, pmx_c),
						   'future_info' : mi_x1x2_c(py, pm, pym_c),
						   'functional' : -np.log2(np.inner(z,px))})
    # among the restarts, select the result that gives the minimum
    # value for the functional we're actually minimizing (eq 29 in
    # Tishby et al 2000).
    selected_candidate = min(candidates, key=lambda c: c['functional'])
    i_p = selected_candidate['past_info']
    i_f = selected_candidate['future_info']
    return [i_p,i_f,b]


def IB(px,py,pyx_c,maxbeta=5,numbeta=30,iterations=100,restarts=3,processes=1):
    """Compute an Information Bottleneck curve
    
    Arguments:
    px -- marginal probability distribution for the past
    py -- marginal distribution for the future
    maxbeta -- the maximum value of beta to use to compute the curve
    iterations -- number of iterations to use to for the curve to converge for each value of beta
    restarts -- number of times the optimization procedure should be restarted (for each value of beta) from different random initial conditions.
    processes -- number of cpu threads to run in parallel (default = 1)
    
    Returns:
    ips -- values of ipast for each beta considered
    ifs -- values of ifuture for each beta considered
    bs -- values of beta considered
    mixy -- mutual information between x and y (curve saturation point)
    hx -- entropy of x (maximum ipast value)
    """

    pm_size = px.size                      #Get compression size 
    bs = np.linspace(0.01,maxbeta,numbeta) #value of beta
    
    #Parallel computing of compression for desired beta values
    pool = mp.Pool(processes=processes)
    results = [pool.apply_async(beta_iter,args=(b,px,py,pyx_c,pm_size,restarts,iterations,)) for b in bs]
    pool.close()
    results = [p.get() for p in results]
    ips = [x[0] for x in results]
    ifs = [x[1] for x in results]
    
    #Values of beta may not be sorted appropriately, code below sorts ipast and ifuture according to their corresponding value of beta, and in correct order
    b_s = [x[2] for x in results]          
    ips = [x for _, x in sorted(zip(b_s,ips))]
    ifs = [x for _, x in sorted(zip(b_s,ifs))]
    
    # restrict the returned values to those that, at each value of
    # beta, actually increase (for Ipast) and do not decrease (for
    # Ifuture) the information with respect to the previous value of
    # beta. This is to avoid confounds from cases where the AB
    # algorithm gets stuck in a local minimum.
    ub, bs = compute_upper_bound(ips, ifs, bs)
    ips = np.squeeze(ub[:,0])
    ifs = np.squeeze(ub[:,1])
    
    # Return saturation point (mixy) and max horizontal axis (hx)
    mixy = mi_x1x2_c(py, px, pyx_c)
    hx = entropy(px,base=2)
    return ips, ifs, bs, mixy, hx


def p_mx_c(pm,px,py,pyx_c,pym_c,beta):
    """Update conditional distribution of bottleneck random variable given x.
    
    Arguments:
    pm -- marginal distribution p(M) - vector
    px -- marginal distribution p(X) - vector
    py -- marginal distribution p(Y) - vector
    pyx_c -- conditional distribution p(Y|X) - matrix
    pym_c -- conditional distribution p(Y|M) - matrix
    
    Returns:
    pmx_c -- conditional distribution p(M|X)
    z -- normalizing factor
    """
    
    pmx_c = np.zeros((pm.size,px.size)) # P(M|X) matrix to be returned
    for mi in range(pm.size):
        for xi in range(px.size):
            pmx_c[mi,xi] = pm[mi] * np.exp(-beta * entropy(pyx_c[:,xi], pym_c[:,mi], base=2))
    z = pmx_c.sum(axis=0)
    pmx_c /= z #Normalize 
    return pmx_c, z


def p_ym_c(pm,px,py,pyx_c,pmx_c):
    """Update conditional distribution of bottleneck variable given y.
    
    Arguments:
    pm -- Marginal distribution p(M)
    px -- marginal distribution p(X)
    pyx_c -- conditional distribution p(Y|X)
    pmx_c -- conditional distribution p(M|X)
    
    Returns:
    pym_c -- conditional distribution p(Y|M)
    """
    pym_c = np.zeros((py.size,pm.size))
    for yi in range(py.size):
        for mi in range(pm.size):
            for xi in range(px.size):
                pym_c[yi,mi] += (1./pm[mi])*pyx_c[yi,xi]*pmx_c[mi,xi]*px[xi]
    return pym_c


def p_m(pmx_c,px):
    """Update marginal distribution of bottleneck variable.
    
    Arguments:
    pmx_c -- conditional distribution p(M|X)
    px -- marginal distribution p(X)
    
    Returns:
    pm - marginal distribution of compression p(M)
    """
    pm = np.zeros(pmx_c.shape[0])
    for mi in range(pm.size):
        for xi in range(px.size):
            pm[mi] += pmx_c[mi,xi]*px[xi]
    return pm



import unittest
import numpy as np

import embo

def test_origin(x,y):
    """Check that the IB bound starts at (0,0) for small beta"""
    i_p,i_f,_ = embo.empirical_bottleneck(x,y)
    np.testing.assert_allclose((i_p[0],i_f[0]),(0,0),rtol=1e-7,atol=1e-9)

def test_asymptote(x,y):
    """Check that the IB bound saturates at (H(x),MI(X:Y)) for large beta.
    
    Note that both H(X) and MI(X,Y) are computed using the functions
    defined within EMBO.

    """
    i_p,i_f,beta,mi,hx,hy = embo.empirical_bottleneck(x,y,maxbeta=10,return_entropies=True)
    np.testing.assert_allclose((i_p[-1],i_f[-1]),(hx,mi),rtol=1e-7)


class TestBinarySequence(unittest.TestCase):
    def setUp(self):
        # Fake data sequence
        self.x = np.array([0,0,0,1,0,1,0,1,0,1])
        self.y = np.array([1,0,1,0,1,0,1,0,1,0])

    def test_origin(self):
        """Check beta->0 limit for binary sequence"""
        test_origin(self.x, self.y)
        
    def test_asymptote(self):
        """Check beta->infinity limit for binary sequence"""
        test_asymptote(self.x, self.y)

class TestUpperBound(unittest.TestCase):
    def setUp(self):
        self.a = np.array([[0,0],[1,1],[2,0],[3,3],[3,4],[2,5],[4,6],[2,7],[3,8]])
        self.betas = np.arange(self.a.shape[0])
        self.true_idxs = np.array([0,1,3,6], dtype=np.int)

    def test_upper_bound(self):
        """Check extraction of upper bound in IB space"""
        u = embo.compute_upper_bound(self.a[:,0], self.a[:,1])
        np.testing.assert_array_equal(u, self.a[self.true_idxs,:])

    def test_betas(self):
        """Check extraction of beta values related to upper bound in IB space"""
        u, betas = embo.compute_upper_bound(self.a[:,0], self.a[:,1], self.betas)
        np.testing.assert_array_equal(betas, self.betas[self.true_idxs])
        
class TestArbitraryAlphabet(unittest.TestCase):
    def setUp(self):
        # Fake data sequence
        self.x = np.array([0,0,0,2,0,2,0,2,0,2])
        self.y = np.array([3.5,0,3.5,0,3.5,0,3.5,0,3.5,0])

    def test_origin(self):
        """Check beta->0 limit for sequence with arbitrary alphabet"""
        test_origin(self.x, self.y)
        
    def test_asymptote(self):
        """Check beta->infinity limit for sequence with arbitrary alphabet"""
        test_asymptote(self.x, self.y)

if __name__ == "__main__":
    unittest.main(verbosity=2)



from __future__ import division

import numpy as np
from scipy.stats import entropy

def p_joint(x1,x2,windowx1=1,windowx2=1):
    """
    Compute the joint distribution between two data series
    
    x1 = first array
    x2 = second array
    windowx1 = moving window size to consider for the x1 array
    windowx2 = moving window size to consider for the x2 array
    
    return a matrix of the joint probability p(x1,x2)
    """
    x1_unique, x1 = np.unique(x1, return_inverse=True)
    x2_unique, x2 = np.unique(x2, return_inverse=True)
    numuniquex1 = x1_unique.size
    numuniquex2 = x2_unique.size
    numwordsx1 = numuniquex1**windowx1
    numwordsx2 = numuniquex2**windowx2
    aux_base_x1 = numuniquex1**np.arange(windowx1)[::-1]
    aux_base_x2 = numuniquex2**np.arange(windowx2)[::-1]
    px1x2 = np.zeros((numwordsx1,numwordsx2)) #matrix of size numwordsx,numwordsy with for the joint probability distribution
    for i in range(len(x1)-(windowx1-1)):
        x1i = np.inner(x1[i:i+windowx1], aux_base_x1).astype(np.int)
        x2i = np.inner(x2[i:i+windowx2], aux_base_x2).astype(np.int)
        px1x2[x1i,x2i] += 1
    return px1x2/px1x2.sum()

def mi_x1x2_c(px1,px2,px1x2_c):
    """Compute the MI between two probability distributions x1 and x2
    using their respective marginals and conditional distribution
    """
    marginal_entropy = entropy(px1, base=2)
    conditional_entropy = 0.
    for x2i in range(px2.size):
        conditional_entropy += px2[x2i] * entropy(px1x2_c[:,x2i], base=2)
    return marginal_entropy - conditional_entropy

def compute_upper_bound(IX, IY, betas=None):
    """Extract the upper part of the convex hull of an IB sequence.

    This is a post-processing step that is needed after computing an
    IB sequence (defined as a sequence of (IX, IY) pairs),
    to remove the random fluctuations in the result induced by the AB
    algorithm getting stuck in local minima.

    Parameters
    ----------
    IX : array
        I(X) values
    IY : array 
        I(Y) values
    betas : array (default None)
        beta values from the IB computation

    Returns
    -------
    array (n x 2)
        (I(X), I(Y)) coordinates of the upper part of the convex hull
        defined by the input points.
    array (n)
        The beta values corresponding to the points of the upper bound.

    """
    points = np.vstack((IX,IY)).T
    selected_idxs = [0]

    for idx in range(1,points.shape[0]):
        if points[idx,0]>points[selected_idxs[-1],0] and points[idx,1]>=points[selected_idxs[-1],1]:
            selected_idxs.append(idx)
            
    upper_bound = points[selected_idxs,:]

    if betas is None:        
        return upper_bound
    else:
        return upper_bound, betas[selected_idxs]




from setuptools import setup
import os
import codecs

NAME = "embo"
VERSION_FILE = "VERSION"
INSTALL_REQUIRES = ["numpy", "scipy"]

here = os.path.abspath(os.path.dirname(__file__))

# get current version
with open(os.path.join(here, NAME, VERSION_FILE)) as version_file:
    VERSION = version_file.read().strip()

# get long description from README
with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

setup (name=NAME,
       version=VERSION,
       url="https://gitlab.com/epiasini/{}".format(NAME),
       description="Empirical Information Bottleneck",
       long_description=LONG_DESCRIPTION,
       long_description_content_type="text/markdown",
       install_requires=INSTALL_REQUIRES,
       python_requires=">=3",
       author="Eugenio Piasini",
       author_email="epiasini@sas.upenn.edu",
       license="GPLv3+",
       classifiers=[
           "Development Status :: 4 - Beta",
           "Intended Audience :: Science/Research",
           "Topic :: Scientific/Engineering :: Information Analysis",
           "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
           "Programming Language :: Python :: 3",
           "Operating System :: OS Independent",
           "Environment :: Console"
       ],
       packages=[NAME,
                 "{}.test".format(NAME)],
       test_suite="{}.test".format(NAME),
       include_package_data=True)
